open Macro
open Str
let debug_flag = false
let dbgmsg header s = 
  if debug_flag then
    prerr_endline (header^": "^s)
  else
    ()
 
module type SIGSERVER =
sig
	type landscape = string
	type ready_set = string		(* excluded prefix *)
	type state = string
	type move = string
	type mapfile = string
	    
	exception Invalid of string 
	  
	val mapsize : int * int 

	val get_ready: mapfile (* -> landscape *) ->  ready_set * ready_set
	val get_ready_set : ready_set * ready_set -> unit
	val proceed: move -> state
	val action: state -> move
	val validcheck: move -> state -> bool 
	val endcheck: state -> bool 
end
 
exception Invalid of string 

(*********** Game Specification *************)
let mapsize = (30,30)
let width = fst mapsize
let height = snd mapsize
let friendnum = 3 				(*	the number of friend	*)
let membernum = friendnum + 1 
(********************************************)

let map = Array.make_matrix (fst mapsize) (snd mapsize) 0
let log = ref ""  
let singcnt = Array.make 10 0
let stopcnt = Array.make 10 0
(* let map2 = ref false *) 
let map3 = ref false
	
module Lmap = Map.Make(struct
  			   type t = int  (* cheolsu friend 1~3 : 1 ~ 3 cheolsu : 4 yungsu friend 1~3 : 5~7 yungsu : 8 yunghee : 9*)
  			   let compare = compare
  			 end)

module Smap = Map.Make(struct
  			   type t = int  
  			   let compare = compare
  			 end)

let lmap = ref (Lmap.empty) 
let smap = ref (Smap.empty)

let record strlst = 
	log := (!log) ^ "\n" ^ (String.concat "\n" strlst)

let tojs n = 
	match n with 
	| 1 -> 0
	| 2 -> 10
	| 3 -> 11
	| 4 -> 12
	| 7 -> 1
	| 9 -> 13
	| _ -> n

let init_get_map mapstr = 
  try
    (		
      let lst' = (Str.split (Str.regexp "[ |\t|\n|\r]+") mapstr) in 
      let lst1 = List.map (fun x -> int_of_string x) lst' in
      let lst2 = List.map (fun x -> (if x = ysbegin then csbegin else if x = csbegin then ysbegin else x)) lst1 in		  
      let yh_x = ref 0 in
      let yh_y = ref 0 in 
	
      (********* Log Record  **********)
      let _ = record ["var cx = "^(string_of_int width)^";"; 
		      "var cy = "^(string_of_int height)^";";
		      "var data = [";
		     ] in
				
      (* configuration for each map *)      
			let _ = if (List.mem chbegin lst1) then map3 := true (* else if (List.mem swamp lst1) then map2 := true *)else () in
			
let _ = 	      
	for i = 0 to (height-1) do
	  for j = 0 to (width-1) do  
	    map.(j).(i) <- (let tile = List.nth lst1 (width * i + j) in
			      if tile = yhbegin then (yh_x := j; yh_y := i; 
						      (lmap := Lmap.add (membernum*2+1) (!yh_x,!yh_y) (!lmap)); flat) 
			      else tile);										
	  done;
	done
      in
	
      let yh_loc = 
	let yhxs = 
	  if (!yh_x) < 10 && (!yh_x >= 0) then 
	    "0"^(string_of_int !yh_x) 
	  else 
	    (string_of_int !yh_x) 
	in 
	let yhys = 
	  if (!yh_y) < 10 && (!yh_y >= 0) then 
	    "0"^(string_of_int !yh_y) 
	  else 
	    (string_of_int !yh_y) 
	in
	  yhxs^yhys
      in
	
      let str1 = ref "" in
      let str2 = ref "" in
      let _ = 
	for i = 0 to (width-1) do
	  for j = 0 to (height-1) do  
	    str1 := !str1 ^ (string_of_int (map.(j).(i)));
	    str2 := !str2 ^ (string_of_int ( 
			       if (map.(j).(i)) = ysbegin then csbegin 
			       else if (map.(j).(i)) = csbegin then ysbegin 
			       else if (!map3) && ((map.(j).(i)) = yubegin) then chbegin 
			       else if (!map3) && ((map.(j).(i)) = chbegin) then yubegin 
			       else (map.(j).(i))));
	  done;
	done in
      let () = dbgmsg "init_get_map len(str1)" (string_of_int (String.length !str1)) in
      let () = dbgmsg "iniy_get_map yh_loc" yh_loc in
	("I"^(!str1)^(yh_loc), "I"^(!str2)^(yh_loc))
	  
    )
  with _ -> raise (Invalid "Invalid map file!")
	
let move_get_map () =	
  let rec mkmemberlst n res =
    if (n > membernum*2+1) then res else mkmemberlst (n+1) (res@[n])
  in
  let memberlst1 = mkmemberlst 1 [] in
  let memberlst2 = List.map (fun x -> if (x <= membernum) then x + membernum else if (x <= membernum*2) then x - membernum else x) memberlst1 in
  let f = (fun x -> let (xl,yl) = Lmap.find x (!lmap) in 
	   let f n = if n < 10 && n >= 0 then ("0"^(string_of_int n)) else (string_of_int n) in 
	     (f xl)^(f yl) )
  in 
  let memberinfo_lst1 = List.map f memberlst1 in 
  let memberinfo_lst2 = List.map f memberlst2 in 
  let resultstr = ref "" in 
    for i = 0 to (width-1) do
      for j = 0 to (height-1) do 				 					
	resultstr := !resultstr ^ (string_of_int (map.(j).(i)))
      done;
    done;	
    
    ("M"^(!resultstr)^(String.concat "" memberinfo_lst1), "M"^(!resultstr)^(String.concat "" memberinfo_lst2))

let get_ready_set (ready1, ready2) = 
  let validloc character (locx, locy) =
    try
      ( 				
	if (character < membernum) then 
	  (map.(locx).(locy) = csbegin)
	else if (character = membernum) then 
	(
	  if (!map3) then (map.(locx).(locy) = chbegin) else (map.(locx).(locy) = csbegin)
	)
	else if (character < membernum*2) then 
	  (map.(locx).(locy) = ysbegin)
	else
	(
	  if (!map3) then (map.(locx).(locy) = yubegin) else (map.(locx).(locy) = ysbegin)
	)
      )
    with _ -> false
  in
  let rec init_set str character = 
    if (character mod membernum = 0) then 
      (
	let (locx, locy) = (int_of_string (String.sub str 0 2), int_of_string (String.sub str 2 2)) in			
	  if (validloc character (locx, locy)) then 
	    (					
	      let _ = map.(locx).(locy) <- flat in                       (*  replace the tile for beginning into flat tile *)					
	      let _ = smap := Smap.add character (0,0) (!smap) in
	      let _ = lmap := Lmap.add character (locx, locy) (!lmap) in
		
		(string_after str 4)
	    )
	  else
	    (
	      if (character <= membernum) then raise (Invalid "Player1 : Invalid init!") else raise (Invalid "Player2 : Invalid init!")
	    )
      )		
    else
      (		
	let validskill (s1,s2) = (s1 >= 1) && (s1 <= 6) && (s2 <= 6) && (s2 >= 1) in
	let (skill1, skill2) = (int_of_string (String.sub str 0 1), int_of_string (String.sub str 1 1)) in 
	let (locx, locy) = (int_of_string (String.sub str 2 2), int_of_string (String.sub str 4 2)) in
	  
	  if (validloc character (locx, locy)) && (validskill (skill1,skill2)) then 
	    (						
	      let _ = map.(locx).(locy) <- flat in                       (*  replace the tile for beginning into flat tile *)	
	      let _ = smap := Smap.add character (skill1, skill2) (!smap) in							
	      let _ = lmap := Lmap.add character (locx, locy) (!lmap) in
	      let _ = if (skill1 = sing) then singcnt.(character) <- singcnt.(character) + 10 in 					
	      let _ = if (skill2 = sing) then singcnt.(character) <- singcnt.(character) + 10 in
		init_set (string_after str 6) (character+1)
	    )
	  else
	    (
	      if (character <= membernum) then raise (Invalid "Player1 : Invalid init!") else raise (Invalid "Player2 : Invalid init!")
	    )			 
      )
  in		
  let _ = smap := Smap.add (membernum*2+1) (0,0) (!smap) in
	  
  let initstr1 = try
    (
      init_set ready1 1  
    ) with _ -> "err"
	in 
	let initstr2 = try
		(
			init_set ready2 (membernum+1)  
   	) with _ -> "err"
	in
	
	let _ = if ((initstr1="err") && (initstr2="err")) then raise (Invalid "Player1 & Player2 : Invalid init!") 
					else if ((initstr1="err") && (not (initstr2="err"))) then raise (Invalid "Player1 : Invalid init!")
					else if ((not (initstr1="err")) && (initstr2="err")) then raise (Invalid "Player2 : Invalid init!")
					else ()
	in
	()

let get_ready file = 
	let rec makeString ic line =
		let buffer = String.create 1000 in
		let count = (input ic buffer 0 1000) in
			if count = 0 then
				""
			else
				line^(String.sub buffer 0 count)^(makeString ic line)
	in

	let ic = open_in file in
	let mapstr' = (makeString ic "") in 
	let _ = close_in ic in 

	let (mapstr1, mapstr2) = (init_get_map mapstr') in 

  let time1 = Sys.time () in  
	let ready_set1 = try ( Client1.Client.get_ready mapstr1 ) with _ -> "err" in	
	let time2 = Sys.time () in
	let ready_set2 = try ( Client2.Client.get_ready mapstr2 ) with _ -> "err" in
	let time3 = Sys.time () in  

  let _ = if ((ready_set1="err") && (ready_set2<>"err")) then raise (Invalid "Player1 : Invalid init!") 
	        else if ((ready_set1<>"err") && (ready_set2="err")) then raise (Invalid "Player2 : Invalid init!")
					else if ((ready_set1="err") && (ready_set2="err")) then raise (Invalid "Player1 & Player2 : Invalid init!")
					else () 
	in 
	
  let _ = if ((time2 -. time1) > 10.0) && ((time3 -. time2) <= 10.0) then raise (Invalid "Player1 : Time over! (init)")
	        else if ((time2 -. time1) <= 10.0) && ((time3 -. time2) > 10.0) then raise (Invalid "Player2 : Time over! (init)")
					else if ((time2 -. time1) > 10.0) && ((time3 -. time2) > 10.0) then raise (Invalid "Player1 & Player2 : Time over! (init)")
					else () 
	in
	(* let _ = record [mapstr1 ; ready_set1 ; ready_set2] in *) 

	get_ready_set (string_after ready_set1 1, string_after ready_set2 1)

let (--) (x1, y1) (x2, y2) = ((x1 - x2) , (y1 - y2)) 

let todirection (x,y) = match (x,y) with 
			| (1,0) -> "\'left\'" 
			| (-1,0) -> "\'right\'"
			| (0,-1) -> "\'down\'"
			| (0,1) -> "\'up\'"
			| _ -> "\'stay\'"

let is_occupied loc lmap =
    Lmap.fold (fun _ char_loc  b -> (char_loc = loc) || b ) lmap false
let rec action actionlst chmap =
    List.iter 
      (fun (character, act, (x,y)) ->
				let (bx,by) = Lmap.find character (!lmap) in 
				(*
				let _ = 	stopcnt.(character) <- stopcnt.(character) + 1;
									dbgmsg "stopcnt : " (string_of_int (stopcnt.(character)));
								if (((Lmap.find character (!lmap)) = (x,y))) && (map.(x).(y) = swamp) then
								( 
									stopcnt.(character) <- stopcnt.(character) + 1;
									dbgmsg "stopcnt : " (string_of_int (stopcnt.(character)))
								) 
								else stopcnt.(character) <- 0 
				in  
				*)
	 let _ = 
		   ( if (x < 0) || (y < 0) then ()	    
	  
		 (* Die in swamp *)
		(*
	   else if (map.(x).(y) = swamp) && (stopcnt.(character) > 1) then 
				let _ = record ["[\'c"^(string_of_int character)^"\', \'die\'], "] in 
				lmap := Lmap.add character (-1,-1) (!lmap)
		 *) 
	   else if (act = shift) then
	   ( 
	       let _ = if (map.(x).(y) = flat) then chmap.(x).(y) <- ((chmap.(x).(y))@[flat]) 
				         else if (map.(x).(y) = swamp) then chmap.(x).(y) <- ((chmap.(x).(y))@[swamp]) 
				 in
				 if (map.(x).(y) = flat) || (map.(x).(y) = swamp) then lmap := Lmap.add character (x,y) (!lmap)				 		  
				 else let _ = record ["[\'c"^(string_of_int character)^"\', \'die\'], "] in
				   lmap := Lmap.add character (-1,-1) (!lmap) 
	   ) 
	   else if (act = rmwall) then chmap.(x).(y) <- if (List.mem flat (chmap.(x).(y))) then chmap.(x).(y) else ((chmap.(x).(y))@[flat])
	   else if (act = mkwall) then chmap.(x).(y) <- if (List.mem wall (chmap.(x).(y))) then chmap.(x).(y) else ((chmap.(x).(y))@[wall])				
	   else if (act = mkpit) then chmap.(x).(y) <- if (List.mem pit (chmap.(x).(y))) then chmap.(x).(y) else ((chmap.(x).(y))@[pit])				
	   else if (act = rmpit) then chmap.(x).(y) <- if (List.mem flat (chmap.(x).(y))) then chmap.(x).(y) else ((chmap.(x).(y))@[flat])				
	   else if (act = mkbrdg) then chmap.(x).(y) <- if (List.mem flat (chmap.(x).(y))) then chmap.(x).(y) else ((chmap.(x).(y))@[flat])
	   else 
		 (
			  (singcnt.(character) <- singcnt.(character) - 1)   (*  Sing  *)				
		 )			  
	 )
	in
	let (cx, cy) = Lmap.find character (!lmap) in
	if (cx, cy) = (-1,-1) then ()
	else
	( 
			let _ = if ((bx,by) <> (cx,cy)) then stopcnt.(character) <- 0 	
				      else if ((bx,by) = (cx,cy)) && ((map.(cx).(cy)) = swamp) then stopcnt.(character) <- stopcnt.(character) + 1
							else ()
			in
			let _ = if (stopcnt.(character) > 1) then  
									let _ = record ["[\'c"^(string_of_int character)^"\', \'die\'], "] in 
											lmap := Lmap.add character (-1,-1) (!lmap)
							else () 
			in
			()
	) 
	
		 
      ) actionlst
	
		  
let validcheck actionlst =
			
  let mem e (e1,e2) = (e=e1) || (e=e2) || (e = 0) in
  let near (x1,y1) (x2,y2) = (abs(x1-x2) + abs(y1-y2)) <= 1 in
    
  let validact act (x,y) =
    let () = dbgmsg "validact: act" (string_of_int act) in
    let () = dbgmsg "validact: (x,y)" 
      ("("^(string_of_int x)^", "^(string_of_int y)^")") in
  try
  (
    if (act = shift) && ((x, y) = (-1, -1)) then true 
    else if (act = shift) then (map.(x).(y) <> wall)  
    else if (act = rmwall) then map.(x).(y) = wall
    else if (act = mkwall) then map.(x).(y) = flat				
    else if (act = mkpit) then map.(x).(y) = flat				
    else if (act = rmpit) then map.(x).(y) = pit				
    else if (act = mkbrdg) then map.(x).(y) = water
    else true   (* sing *) 
  )
  with _ -> false 
 
  in		
    
  let lst = actionlst in   
  let lst' = 
    (List.filter (
       fun (character, act, (x,y)) -> 
	 (
	   (* let _ = prerr_int character ;prerr_int act ; prerr_endline "" in *)
   	   (*  Does he have this skill?? *)
	   let has_skill = (mem act (Smap.find character (!smap))) in
	     (*  Is the target tile near him ?? *)  
	   let is_near = near (x,y) (Lmap.find character (!lmap)) in
	     (* Does the action make sense ?? *)			
	   let is_valid_act = validact act (x,y) in
	     (*		      prerr_int character; prerr_int act; prerr_newline ();
			      prerr_endline (string_of_bool has_skill); *)
	     has_skill && is_near && is_valid_act
	 )
     ) actionlst)
  in
    (*prerr_endline ((string_of_int (List.length lst))^" "^(string_of_int (List.length lst')));*)
    (List.length lst) = (List.length lst')
     
(* let string_of_fvector (x,y) = Printf.sprintf "(%f,%f)" x y *)
let string_of_fvector (x,y) = "("^(string_of_float x)^", "^(string_of_float y)^")"

let singing singers = 
  if (List.length singers) = 0 then (Lmap.find (membernum*2+1) (!lmap)) 
	else
	(
		let (lv, rv, uv, dv) = ((-1.0, 0.0) , (1.0, 0.0), (0.0, 1.0) , (0.0 , -1.0)) in 
		let (yh_x, yh_y) = Lmap.find (membernum*2+1) (* young hee *) (!lmap) in
		
		let size (x, y) = sqrt((x ** 2.0) +. (y ** 2.0)) in 
	 	
		let (++) (x1, y1) (x2, y2) = ((x1 + x2) , (y1 + y2)) in
		let (%) (x1, y1) (x2, y2) = (x1 *. x2) +. (y1 *. y2) in	
						
		let norm (x, y) =
			let (yh_x , yh_y , x , y) = (float_of_int yh_x , float_of_int yh_y , float_of_int x , float_of_int y) in
			let dist2 = ((x -. yh_x) ** 2.0) +. ((y -. yh_y) ** 2.0) in
			let result = 
                if dist2 = 0. then (0.,0.)
                else ((x -. yh_x) /. dist2 , (y -. yh_y) /. dist2) in
            let () = dbgmsg "singing: norm(x,y)" (string_of_fvector result) in
                result
		in  
		let vectors = List.map (fun (character, act, (x,y)) -> norm (Lmap.find character (!lmap))) singers in 
		let sumvector =
            if List.mem (0.,0.) vectors then (0.,0.) 
            else 
            List.fold_left (fun (x1,y1) (x2,y2) -> ((x1 +. x2), (y1 +. y2))) (0.0 , 0.0) vectors in 
        let () = dbgmsg "singing: sumvector" (string_of_fvector sumvector) in
        let (dirx,diry) = 
          if sumvector = (0.,0.) then (0.,0.)
          else (
            let rads = List.map (fun vec -> (vec, abs_float(acos ((vec % sumvector) /.  ((size vec) *. (size sumvector)))))) [lv;uv;rv;dv] in
            let dbg_result = List.map (fun (fvec,f) -> 
              (string_of_fvector fvec)^", "^(string_of_float f)) rads in
            let () = List.iter (fun s -> dbgmsg "singing: rads" s) dbg_result in
		    let comp v1 v2 = compare (snd v1) (snd v2) in 
		      fst (List.hd (List.sort comp rads))
          ) in 
        let () = dbgmsg "singing: (dirx,diry)" (string_of_fvector (dirx,diry)) in
		(* Young hee move *)		
		let (newx, newy) = (yh_x, yh_y) ++ (int_of_float dirx, int_of_float diry) in
		if (map.(newx).(newy) = flat) then (newx, newy) else (Lmap.find (membernum*2+1) (!lmap)) 
	)
		
let proceed () = 
  let beforel = (!lmap) in 
    
  let toskillname skill = 
    if (skill = shift) then ""
    else if (skill = rmwall) then "\'wallremove\', "
    else if (skill = mkwall) then "\'wallcreate\', "
    else if (skill = mkpit) then "\'trapcreate\', "
    else if (skill = rmpit) then "\'trapremove\', "
    else if (skill = mkbrdg) then "\'pondremove\', "
    else "\'sing\', "
  in
    
  let rec action_js lst =			
    match lst with 
	(character, skill, (x,y)) :: tl ->				  
	  (
	    let (beforex, beforey) = Lmap.find character beforel in
	      if ((skill = shift) && (compare (beforex,beforey) (x,y)) = 0)
		|| ((skill = sing) && (singcnt.(character) < 0))
	      then action_js tl				
	      else
		let _ = record ["[\'c"^(string_of_int character)^"\', "
				^toskillname(skill)
				^todirection((beforex,beforey) -- (x,y))^"],"]
		in
		  action_js tl
	  )
      | [] -> ()			 
  in
    
  let is_effective (character, skill, (x,y)) = 
    if (x < 0) || (y < 0) then true 
    else if (skill = shift) then true
    else if (skill = rmwall) then (map.(x).(y) = flat)
    else if (skill = mkwall) then (map.(x).(y) = wall)
    else if (skill = mkpit) then (map.(x).(y) = pit)
    else if (skill = rmpit) then (map.(x).(y) = flat)
    else if (skill = mkbrdg) then (map.(x).(y) = flat)
    else true     (*   TODO  *)
  in
    
  let (state1, state2) = move_get_map () in
	let time1 = Sys.time () in 
  let movestr1 = try ( Client1.Client.action state1 ) with _ -> "err" in
	let time2 = Sys.time () in  
  let movestr2 = try ( Client2.Client.action state2 ) with _ -> "err" in
	let time3 = Sys.time () in  
  
	let _ = if ((movestr1="err") && (movestr2<>"err")) then raise (Invalid ("Player1 : Invalid action! "^movestr1)) 
	        else if ((movestr1<>"err") && (movestr2="err")) then raise (Invalid ("Player2 : Invalid action! "^movestr2))
					else if ((movestr1="err") && (movestr2="err")) then raise (Invalid ("Player1 & Player2 : Invalid action! "^movestr1^" "^movestr2))
					else () 
	in
		  
  (*let _ = record [state1 ; movestr1 ; movestr2] in*) 
  let _ = if ((time2 -. time1) > 10.0) && (time3 -. time2) <= 10.0 then raise (Invalid "Player1 : Time over! (action)")
	        else if ((time2 -. time1) <= 10.0) && (time3 -. time2) > 10.0 then raise (Invalid "Player2 : Time over! (action)")
					else if ((time2 -. time1) > 10.0) && (time3 -. time2) > 10.0 then raise (Invalid "Player1 & Player2 : Time over! (action)")
					else () 
	in
		  
  let chmap = Array.make_matrix (fst mapsize) (snd mapsize) [] in
    
  let _ = 
    for i = 0 to (height-1) do
      for j = 0 to (width-1) do 
	chmap.(j).(i) <- [map.(j).(i)];
      done;
    done			
  in
    
  let rec move_set str character res = 		  
    if (character mod membernum = 0) then 
      (			
	let targetx = int_of_string (String.sub str 0 2) in
	let targety = int_of_string (String.sub str 2 2) in
	let res' = res@[(character , 0 ,  (targetx, targety))] in 
	  
	  res'
      )		
    else
      (			
	let skillno = int_of_string (String.sub str 0 1) in
	let targetx = int_of_string (String.sub str 1 2) in
	let targety = int_of_string (String.sub str 3 2) in				
	let res' = res@[(character , skillno , (targetx, targety))] in
	  
	  move_set (string_after str 5) (character + 1) res'
      )
  in	
  let actionlst1 = try
    		( 
    move_set (string_after movestr1 1) 1 []			
      		) with _ -> []
  in
    
  let actionlst2 = try
    (
      move_set (string_after movestr2 1) (membernum+1) []			  
    ) with _ -> []
  in
  
	let _ = if (List.length actionlst1 = 0) && (List.length actionlst2 > 0) then raise (Invalid ("Player1 : Invalid action! "^movestr1)) 
				  else if (List.length actionlst1 > 0) && (List.length actionlst2 = 0) then raise (Invalid ("Player2 : Invalid action! "^movestr2)) 
					else if (List.length actionlst1 = 0) && (List.length actionlst2 = 0) then raise (Invalid ("Player1 & Player2 : Invalid action! "^movestr1^" "^movestr2))
					else ()
	in
	
	let (valid1, valid2) = (validcheck actionlst1, validcheck actionlst2) in 	  
  let _ = if ((not valid1) && valid2) then raise (Invalid ("Player1 : Invalid action! "^movestr1))
					else if (valid1 && (not valid2)) then raise (Invalid ("Player2 : Invalid action! "^movestr2))
	 				else if ((not valid1) && (not valid2)) then raise (Invalid ("Player1 & Player2 : Invalid action! "^movestr1^" "^movestr2))
					else (action actionlst1 chmap); action actionlst2 chmap
	in
		
  let singers = List.filter (fun (character, act, (x,y)) -> (act = sing) && (singcnt.(character) >= 0)) (actionlst1@actionlst2) in 
  let (newx, newy) = singing singers in
  let (beforex, beforey) = Lmap.find (membernum*2+1) (!lmap) in 

  (********* Log Record *********)
  let _ = if ((compare (newx,newy) (beforex, beforey)) <> 0) then 
    record ["[\'c0\', "^todirection((Lmap.find (membernum*2+1) (!lmap)) -- (newx,newy))^"],"]
  else ()
  in
    
  let _ = lmap := Lmap.add (membernum*2+1) (newx,newy) (!lmap) in	 
  let _ = chmap.(newx).(newy) <- (chmap.(newx).(newy))@[flat] in  
  let _ = 	
    for i = 0 to (height-1) do
      for j = 0 to (width-1) do
	let chtiles = chmap.(j).(i) in 
	  map.(j).(i) <- if (List.length chtiles) = 2 then (
        let next_tile = List.nth chtiles 1 in
        if (is_occupied (j,i) !lmap) && ((next_tile = wall) || (next_tile = pit)) then 
          List.hd chtiles
        else
          List.nth chtiles 1 
      ) else List.hd chtiles 
      done;
    done
  in  
    
    (********* Log Record *********)
    action_js (List.filter is_effective (actionlst1@actionlst2))
				
	 
let endcheck () =  
  let csloc = Lmap.find membernum (!lmap) in   
  let ysloc = Lmap.find (membernum*2) (!lmap) in
  let yhloc = Lmap.find (membernum*2+1) (!lmap) in
    
  let cswin = (csloc = yhloc) || (ysloc = (-1,-1)) in 
  let yswin = (ysloc = yhloc) || (csloc = (-1,-1)) in
    
    if (cswin && yswin) then
      let csfriend_num = List.length (List.filter (fun x -> (Lmap.find x (!lmap)) <> (-1,-1)) [1;2;3]) in 
      let ysfriend_num = List.length (List.filter (fun x -> (Lmap.find x (!lmap)) <> (-1,-1)) [5;6;7]) in 
	(cswin && (csfriend_num >= ysfriend_num) , yswin && (ysfriend_num >= csfriend_num))				
    else
      (cswin, yswin)  	

let rec output_str dest result i lim =
        if (i < lim) then
                let _ = output_char dest result.[i] in
                        output_str dest result (i+1) lim
        else
                ()
let strlen str =
        (String.index_from str 0 '\000')							
		
let main () =		
	let dest = open_out "data.js" in
	let jsout dest result =
		let result = result ^ "\000" in
     output_str dest result 0 (strlen result);		        
     close_out dest		
	in
  try
    (
      let mapfile = Sys.argv.(1) in		
      let _ = (if (Sys.file_exists mapfile) then () else (print_string("no file exist!\n"); exit (-1))) in
	
      let win = ref (false, false) in
      let turn = ref 0 in 
      let no_win (cswin, yswin) = (not cswin) && (not yswin) in 
      let temp = ref "" in 
      let charac = ref false in
	get_ready mapfile;
	for i = 0 to (height-1) do
	  for j = 0 to (width-1) do
	    for k = 1 to membernum*2+1 do						 
	      if ((j,i) = (Lmap.find k (!lmap))) then
		(  
		  (********* Log Record  **********)
		  temp := !temp ^ ((string_of_int (if k = membernum*2+1 then 1 else k+1))^", ");
		  charac := true;
		)						
	    done;
	    if (!charac) then
	      (
		charac := false;						
	      )
	    else
	      temp := !temp ^ ((string_of_int (tojs map.(j).(i)))^", "); 
	  done;
	  record [!temp];
	  temp := "";
	done;  
	record ["];";
		"var scenario = [";
	       ];
	
	while ((no_win !win) && (!turn < 100)) do 
	  record ["["];					 
	  proceed ();
	  record ["],"]; 
	  win := endcheck ();
	  incr turn;
	  (*
	    List.iter (fun (a,b) -> print_endline ((string_of_int a)^", "^(string_of_int b))) 
	    (List.map (fun x -> Lmap.find x (!lmap)) [1;2;3;4;5;6;7;8;9]) ;
	  *)
	done;  
	
  
	match (!win) with  
	  | (true, false) -> let _ = record ["[[\'message\',\'show\' ,\'Player1(Cheol-su) wins!\']]];"] in jsout dest (!log); exit(51)
	  | (false, true) -> let _ = record ["[[\'message\',\'show\' , \'Player2(Yeong-su) wins!\']]];"] in jsout dest (!log); exit(52)
	  | _ ->  let _ = record ["[[\'message\', \'show\',\'Draw!\']]];"] in jsout dest (!log); exit(53)
    )
  with Invalid s -> let _ = record ["[\'message\', \'show\',\'"^s^"\']]];"] in 
										let _ = jsout dest (!log) in										
													 if (String.compare s "Player1 : Invalid init!") = 0 then exit (54)
											else if (String.compare s "Player2 : Invalid init!") = 0 then exit (55)
											else if (String.compare s "Player1 & Player2 : Invalid init!") = 0 then exit (56)
											else if (String.compare (String.sub s 0 25) "Player1 : Invalid action!") = 0 then exit (57)
										  else if (String.compare (String.sub s 0 25) "Player2 : Invalid action!") = 0 then exit (58)
											else if (String.compare (String.sub s 0 35) "Player1 & Player2 : Invalid action!") = 0 then exit (59)
											else if (String.compare s "Player1 : Time over! (init)") = 0 then exit (60)
											else if (String.compare s "Player2 : Time over! (init)") = 0 then exit (61)
											else if (String.compare s "Player & Player2 : Time over! (init)") = 0 then exit (62)
											else if (String.compare s "Player1 : Time over! (action)") = 0 then exit (63)
											else if (String.compare s "Player2 : Time over! (action)") = 0 then exit (64)
											else if (String.compare s "Player1 & Player2 : Time over! (action)") = 0 then exit (65)
											else exit(66)
		
let _ = main()		 
		
	
